/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2018     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: support_emusb@segger.com         *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Device * USB Device stack for embedded applications    *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Device version: V3.16a                                 *
*                                                                    *
**********************************************************************
----------------------------------------------------------------------
File        : IP_Config_RNDIS.c
Purpose     : Configuration file for embOS/IP via USB RNDIS.

              The Remote Network Driver Interface Specification (RNDIS)
              is a Microsoft USB class protocol which can be used to
              create a virtual Ethernet connection between a USB
              device and a host PC.

              Any IP_* sample application can be used with this
              configuration to run via USB.

              This configuration enables NetBIOS, which allows
              the user to reach the target via a plain text name.
              Notes:
                (1): This configuration file does not apply any MAC
                     addresses to the interfaces.
                     These interfaces are TARGET - embOS/IP interface
                     and the HOST - RNDIS network card
                     The MAC addresses are automatically generated
                     by the internal driver.
---------------------------END-OF-HEADER------------------------------
*/

#include "BSP.h"
#include "RTOS.h"
#include "IP.h"
#include "USB.h"
#include "USB_RNDIS.h"
#include "USB_Driver_embOS_IP.h"

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/
#define ALLOC_SIZE                 0x6000                      // Size of memory dedicated to the stack in bytes
#define DRIVER                     &USB_IP_Driver              // Driver used for target.

#define _USB_VENDOR_ID             0x8765
#define _USB_PRODUCT_ID            0x3000
#define _USB_VENDOR_NAME           "SEGGER"
#define _USB_PRODUCT_NAME          "SEGGER RNDIS device"
#define _USB_SERIAL_NO             "13245678"

#define _USB_RNDIS_VENDOR_ID       0x0022C7      // 24-bit Organizationally Unique Identifier (OUI) of the vendor.
                                                 // This is the same value as the one stored in the first 3 bytes of
                                                 // a HW (MAC) address.
#define _USB_RNDIS_DESCRIPTION    "SEGGER RNDIS device"
#define _USB_RNDIS_DRIVER_VERSION 0x0100        // 16-bit value representing the software version. The high-order byte specifies
                                                // the major version and the low-order byte the minor version.

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/
#ifdef __ICCARM__          // IAR ARM compiler
  #define PRAGMA(x)         _Pragma(#x)
  #define SECTION(x)                  PRAGMA(location=x)
  #define SECTION_NO_INIT(x)          PRAGMA(location=x)   __no_init
  #define SECTION_NO_INIT_STATIC(x)   PRAGMA(location=x)   static __no_init
#elif (defined __CC_ARM)   // Keil/ARM compiler
  #define SECTION(x)                  __attribute__ ((section (x)))
  #define SECTION_NO_INIT(x)          __attribute__ ((section (x), zero_init))
  #define SECTION_NO_INIT_STATIC(x)   __attribute__ ((section (x), zero_init)) static
  #define ___WEAK                     __attribute__ ((weak))
#elif (defined __GNUC__)   // GNU ARM compiler
  #define SECTION(x)                  __attribute__ ((section (x)))
  #define SECTION_NO_INIT(x)          __attribute__ ((section (x)))
  #define SECTION_NO_INIT_STATIC(x)   __attribute__ ((section (x))) static
#else
  #define SECTION(x)
  #define SECTION_NO_INIT(x)
  #define SECTION_NO_INIT_STATIC(x)   static
#endif

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

/*********************************************************************
*
*       Information that are used during enumeration
*/
const USB_DEVICE_INFO USB_DeviceInfo = {
  _USB_VENDOR_ID,
  _USB_PRODUCT_ID,
  _USB_VENDOR_NAME,
  _USB_PRODUCT_NAME,
  _USB_SERIAL_NO
};

const USB_RNDIS_DEVICE_INFO USB_RNDIS_DeviceInfo = {
  _USB_RNDIS_VENDOR_ID,
  _USB_RNDIS_DESCRIPTION,
  _USB_RNDIS_DRIVER_VERSION
};
//
// USB RNIDS related variables
//
SECTION_NO_INIT_STATIC("USB_RAM")
U8 _abReceiveBuffer[USB_HS_BULK_MAX_PACKET_SIZE * 2];   // Buffer to store incoming data from host.
static unsigned   _IFaceId;

//
// IP related
//
SECTION_NO_INIT_STATIC("IP_RAM")
U32 _aPool[ALLOC_SIZE / 4];                  // This is the memory area used by the stack
//
// NetBIOS name table.
//
const IP_NETBIOS_NAME _aNetNames[] = {
  {"SEGGER.RNDIS", 12},
  {0       , 0}
};

/*********************************************************************
*
*       Static code
*
**********************************************************************
*/
/*********************************************************************
*
*       _AddRNDIS
*
*  Function description
*    Adds and configures RNDIS communication device class to USB stack.
*    See Notes(1).
*/
static void _AddRNDIS(void) {
  USB_RNDIS_INIT_DATA InitData = {0};

  memset(&InitData, 0, sizeof(InitData));
  InitData.EPOut      = USBD_AddEP(USB_DIR_OUT, USB_TRANSFER_TYPE_BULK, 0, _abReceiveBuffer, sizeof(_abReceiveBuffer));
  InitData.EPIn       = USBD_AddEP(USB_DIR_IN,  USB_TRANSFER_TYPE_BULK, 0, NULL, 0);
  InitData.EPInt      = USBD_AddEP(USB_DIR_IN,  USB_TRANSFER_TYPE_INT, 32, NULL, 0);
  InitData.pDriverAPI = &USB_Driver_IP_NI;
  InitData.DriverData.pDriverData = (void *)_IFaceId;
  USBD_RNDIS_Add(&InitData);
}

/*********************************************************************
*
*       _Connect()
*
* Function description
*   Performs a first time initialization for the USB RNDIS.
*
* Parameters
*   IFaceId: Zero-based interface index.
*
* Return value
*   O.K. :   0.
*   Error: < 0.
*/
static int _Connect(unsigned IFaceId) {
  U32 ServerIpAddr;
  U32 SubnetMask;

  ServerIpAddr = IP_BYTES2ADDR(10, 0, 0, 10);
  SubnetMask = IP_BYTES2ADDR(0xff,0xff,0xff,0xf8);
  IP_SetAddrMaskEx(IFaceId, ServerIpAddr, SubnetMask);
  IP_DHCPS_ConfigPool(IFaceId, ServerIpAddr + 1, SubnetMask, 4);  // Setup IP pool to distribute.
  IP_DHCPS_ConfigDNSAddr(IFaceId, &ServerIpAddr, 1);
  IP_DHCPS_Init(IFaceId);
  IP_DHCPS_Start(IFaceId);
  IP_NETBIOS_Init(IFaceId, _aNetNames, 0);                             // Init NetBIOS.
  IP_NETBIOS_Start(IFaceId);                                           // Start NetBIOS.
  USBD_Init();
  USBD_SetDeviceInfo(&USB_DeviceInfo);
  USBD_RNDIS_SetDeviceInfo(&USB_RNDIS_DeviceInfo);
  //
  // Although we do not have a composite device, we enable IAD as a workaround
  // for the buggy Intel USB driver on Windows 7
  //
  USBD_EnableIAD();
  _AddRNDIS();
  USBD_Start();
  return 0;                   // Successfully connected.
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       IP_X_Config
*
*  Function description
*    This function is called by the IP stack during IP_Init().
*
*  Typical memory/buffer configurations:
*    Microcontroller system, size optimized
*      #define ALLOC_SIZE 0x3000                    // 12KBytes RAM
*      mtu = 576;                                   // 576 is minimum acc. to RFC, 1500 is max. for Ethernet
*      IP_SetMTU(0, mtu);                           // Maximum Transmission Unit is 1500 for ethernet by default
*      IP_AddBuffers(8, 256);                       // Small buffers.
*      IP_AddBuffers(4, mtu + 16);                  // Big buffers. Size should be mtu + 16 byte for ethernet header (2 bytes type, 2*6 bytes MAC, 2 bytes padding)
*      IP_ConfTCPSpace(1 * (mtu-40), 1 * (mtu-40)); // Define the TCP Tx and Rx window size
*
*    Microcontroller system, speed optimized or multiple connections
*      #define ALLOC_SIZE 0x6000                    // 24 KBytes RAM
*      mtu = 1500;                                  // 576 is minimum acc. to RFC, 1500 is max. for Ethernet
*      IP_SetMTU(0, mtu);                           // Maximum Transmission Unit is 1500 for ethernet by default
*      IP_AddBuffers(12, 256);                      // Small buffers.
*      IP_AddBuffers(6, mtu + 16);                  // Big buffers. Size should be mtu + 16 byte for ethernet header (2 bytes type, 2*6 bytes MAC, 2 bytes padding)
*      IP_ConfTCPSpace(3 * (mtu-40), 3 * (mtu-40)); // Define the TCP Tx and Rx window size
*
*    System with lots of RAM
*      #define ALLOC_SIZE 0x20000                   // 128 KBytes RAM
*      mtu = 1500;                                  // 576 is minimum acc. to RFC, 1500 is max. for Ethernet
*      IP_SetMTU(0, mtu);                           // Maximum Transmission Unit is 1500 for ethernet by default
*      IP_AddBuffers(50, 256);                      // Small buffers.
*      IP_AddBuffers(50, mtu + 16);                 // Big buffers. Size should be mtu + 16 byte for ethernet header (2 bytes type, 2*6 bytes MAC, 2 bytes padding)
*      IP_ConfTCPSpace(5 * (mtu-40), 5 * (mtu-40)); // Define the TCP Tx and Rx window size
*/
void IP_X_Config(void) {
  int mtu;
  int IFaceId;

  IP_AssignMemory(_aPool, sizeof(_aPool));         // Assigning memory should be the first thing
  IFaceId = IP_AddEtherInterface(DRIVER);          // Add driver for your hardware.
  IP_SetIFaceConnectHook(IFaceId, _Connect);
  _IFaceId = IFaceId;
  //
  // Run-time configure buffers.
  // The default setup will do for most cases.
  //
  mtu = 1500;                                  // 576 is minimum acc. to RFC, 1500 is max. for Ethernet
  IP_SetMTU(IFaceId, mtu);                     // Maximum Transmission Unit is 1500 for ethernet by default
  IP_AddBuffers(12, 256);                      // Small buffers.
  IP_AddBuffers( 9, mtu + 16);                 // Big buffers. Size should be mtu + 16 byte for ethernet header (2 bytes type, 2*6 bytes MAC, 2 bytes padding)
  IP_ConfTCPSpace(3 * (mtu-40), 3 * (mtu-40)); // Define the TCP Tx and Rx window size
  IP_SOCKET_SetDefaultOptions(0
//                              | SO_TIMESTAMP   // Send TCP timestamp to optimize the round trip time measurement. Normally not used in LAN.
                              | SO_KEEPALIVE   // Enable keepalives by default for TCP sockets.
                             );
  //
  // Define log and warn filter.
  // Note: The terminal I/O emulation might affect the timing of your
  //       application, since most debuggers need to stop the target
  //       for every terminal I/O output unless you use another
  //       implementation such as DCC or SWO.
  //
  IP_SetWarnFilter(0xFFFFFFFF);              // 0xFFFFFFFF: Do not filter: Output all warnings.
  IP_SetLogFilter(0
                  | IP_MTYPE_APPLICATION     // Output application messages.
                  | IP_MTYPE_INIT            // Output all messages from init.
                  | IP_MTYPE_LINK_CHANGE     // Output a message if link status changes.
                  | IP_MTYPE_PPP             // Output all PPP/PPPoE related messages.
                  | IP_MTYPE_DHCP            // Output general DHCP status messages.
#if IP_SUPPORT_IPV6
                  | IP_MTYPE_IPV6            // Output IPv6 address related messages
#endif
//                  | IP_MTYPE_DHCP_EXT        // Output additional DHCP messages.
//                  | IP_MTYPE_CORE            // Output log messages from core module.
//                  | IP_MTYPE_ALLOC           // Output log messages for memory allocation.
//                  | IP_MTYPE_DRIVER          // Output log messages from driver.
//                  | IP_MTYPE_ARP             // Output log messages from ARP layer.
//                  | IP_MTYPE_IP              // Output log messages from IP layer.
//                  | IP_MTYPE_TCP_CLOSE       // Output a log messages if a TCP connection has been closed.
//                  | IP_MTYPE_TCP_OPEN        // Output a log messages if a TCP connection has been opened.
//                  | IP_MTYPE_TCP_IN          // Output TCP input logs.
//                  | IP_MTYPE_TCP_OUT         // Output TCP output logs.
//                  | IP_MTYPE_TCP_RTT         // Output TCP round trip time (RTT) logs.
//                  | IP_MTYPE_TCP_RXWIN       // Output TCP RX window related log messages.
//                  | IP_MTYPE_TCP             // Output all TCP related log messages.
//                  | IP_MTYPE_UDP_IN          // Output UDP input logs.
//                  | IP_MTYPE_UDP_OUT         // Output UDP output logs.
//                  | IP_MTYPE_UDP             // Output all UDP related messages.
//                  | IP_MTYPE_ICMP            // Output ICMP related log messages.
//                  | IP_MTYPE_NET_IN          // Output network input related messages.
//                  | IP_MTYPE_NET_OUT         // Output network output related messages.
//                  | IP_MTYPE_DNS             // Output all DNS related messages.
//                  | IP_MTYPE_SOCKET_STATE    // Output socket status messages.
//                  | IP_MTYPE_SOCKET_READ     // Output socket read related messages.
//                  | IP_MTYPE_SOCKET_WRITE    // Output socket write related messages.
//                  | IP_MTYPE_SOCKET          // Output all socket related messages.
                 );
  //
  // Add protocols to the stack.
  //
  IP_TCP_Add();
  IP_UDP_Add();
  IP_ICMP_Add();
#if IP_SUPPORT_IPV6
  IP_IPV6_Add(IFaceId);
#endif
}

/*************************** End of file ****************************/
